/*
 * arduinoSerial
 * -----------------
 * Purpose: translate serial commands into Arduino languague commands,
 * mostly relating to IO.  This would allow a computer to easily take
 * advantage of Arduino's great IO capabilities, while cpu number crunching
 * could be done on the computer (e.g. Video processing)
 *
 * Created 2 Februrary 2009
 * copyleft 2009 Greg 'GroG' Perry <greg.perry@myrobotlab.com>
 * http://myrobotlab.com/
 * 
 */

// TODO - time to upgrade to 017+ http://www.arduino.cc/en/Reference/Servo vs http://www.arduino.cc/playground/ComponentLib/Servo
// TODO - max servo test

#include <avr/interrupt.h>
#include <Servo.h>

#define DIGITAL_WRITE        0 
#define DIGITAL_VALUE        1 
#define ANALOG_WRITE         2 
#define ANALOG_VALUE         3  
#define PINMODE              4
#define PULSE_IN             5 
#define SERVO_ATTACH         6 
#define SERVO_WRITE          7 
#define SERVO_SET_MAX_PULSE  8 
#define SERVO_DETACH         9
#define SET_PWM_FREQUENCY    11
#define SERVO_READ           12
#define ANALOG_READ_POLLING_START	 13
#define ANALOG_READ_POLLING_STOP	 14
#define DIGITAL_READ_POLLING_START	 15
#define DIGITAL_READ_POLLING_STOP	 16

Servo servos[2]; // Lame 2?
//Servo servos[MAX_SERVOS]; TODO - difference between Servo.h SoftwareServo.h MegaServo.h ????

unsigned long loopCount = 0;
int byteCount 		= 0;
int newByte 		= 0;
int ioCommand[3]; 	// most io fns can cleanly be done with a 3 byte code
int readValue; 

int digitalReadPin[13];          	// array of pins to read from
int digitalReadPollingPinCount = 0;     // number of pins currently reading
int lastDigitalInputValue[13];   	// array of last input values

int analogReadPin[4];          		// array of pins to read from
int analogReadPollingPinCount = 0;     	// number of pins currently reading
int lastAnalogInputValue[4];   		// array of last input values

unsigned long retULValue;  
int prescalerVal = 0;
int sensePin = 2;

/*  TODO ISR 
// We need to declare the data exchange
// variable to be volatile - the value is
// read from memory.
volatile int value = 0;

// Install the interrupt routine.
ISR(INT0_vect) {
  // check the value again - since it takes some time to
  // activate the interrupt routine, we get a clear signal.
  value = digitalRead(sensePin);
}
*/

void setup() {
  //First clear all three prescaler bits:
//  prescalerVal = 0x07; //create a variable called prescalerVal and set it equal to the binary                                                       number "00000111"
//  TCCR0B &= ~prescalerVal; //AND the value in TCCR0B with binary number "11111000"

  //Now set the appropriate prescaler bits:
//  int prescalerVal = 1; //set prescalerVal equal to binary number "00000001"
//  TCCR0B |= prescalerVal; //OR the value in TCCR0B with binary number "00000001"

  //pinMode(ledPin,OUTPUT);    // declare the LED's pin as output
  // TODO init all as OUTPUT ! - or find documentation that say
  Serial.begin(9600);        // connect to the serial port    
  //Serial.begin(57600);        // connect to the serial port    
  
//  analogReadPin[analogReadPollingPinCount] = 0; // put on polling read list
//  ++analogReadPollingPinCount;
  
}

void setPWMFrequency (int address, int prescalar)
{
  int clearBits = 0x07;
  if (address == 0x25)
  {
    TCCR0B &= ~clearBits; 
    TCCR0B |= prescalar;
  } else if (address == 0x2E)
  {
    TCCR1B &= ~clearBits; 
    TCCR1B |= prescalar;    
  } else if (address == 0xA1)
  {
    TCCR2B &= ~clearBits; 
    TCCR2B |= prescalar;    
  }
  
}


void removeAndShift (int array [], int& len, int removeValue)
{
  int pos = -1;

  if (len == 0)
  {
	  return;
  }

  // find position of value
  for (int i = 0; i < len; ++i)
  {
	if (removeValue == array[i])
	{
	  pos = i;
	  break;
	}
  }
  // if at the end just decrement size
  if (pos == len - 1)
  {
	  --len;
	  return;
  }
	  
  // if found somewhere else shift left
  if (pos < len && pos > -1)
  {
	  for (int j = pos; j < len - 1; ++j)
	  {
		array[j] = array[j+1];
	  }
	  --len;
  }
}


/*
set pollingRead = true if any of the pins are commanded to read
default behavior should be to poll read
*/
char dbg[4];

void loop () {

  ++loopCount;
  
  // handle serial data begin 
  if (Serial.available () > 0)
  {
      // read the incoming byte:
      newByte = Serial.read();
      ioCommand[byteCount] = newByte;
      byteCount++;

      // command & return language (3 byte protocol) FUNCTION|DATA0|DATA1
      // digitalWrite (13, 1)     = DIGITAL_WRITE|13|1 = 0|13|1
      // results of analogRead(3) = ANALOG_READ|3|253  = 3|3|253
      if (byteCount > 2)
      {
        switch (ioCommand[0])
        {
        
           /* diagnostic echo
           Serial.print(ioCommand[0]);
           Serial.print(ioCommand[1]);
           Serial.print(ioCommand[2]);
           */
           
           case DIGITAL_WRITE:
             digitalWrite(ioCommand[1], ioCommand[2]);
           break;
/*           
           case DIGITAL_READ:
             Serial.write(DIGITAL_READ);
             Serial.write(ioCommand[1]);
             Serial.write(digitalRead(ioCommand[1]));
           break;
*/           
           case ANALOG_WRITE:
             analogWrite(ioCommand[1], ioCommand[2]);
           break;
/*           
           case ANALOG_READ:
             Serial.write(ANALOG_READ);
             Serial.write(ioCommand[1]);
             Serial.write(analogRead(ioCommand[1]));
           break;
*/           
           case PINMODE:
             pinMode(ioCommand[1], ioCommand[2]);
           break;
           case PULSE_IN:
             retULValue = pulseIn(ioCommand[1], ioCommand[2]);
           break;
           case SERVO_ATTACH:
             servos[ioCommand[1]].attach(ioCommand[2]);
           break;
           case SERVO_WRITE:
             servos[ioCommand[1]].write(ioCommand[2]);          
           break;
           case SERVO_READ:
             Serial.write(SERVO_READ);
             Serial.write(ioCommand[1]);
             Serial.write(servos[ioCommand[1]].read());          
           break;
           case SERVO_SET_MAX_PULSE:
             //servos[ioCommand[1]].setMaximumPulse(ioCommand[2]);    TODO - lame fix hardware
           break;
           case SERVO_DETACH:
             servos[ioCommand[1]].detach();
           break;
           case SET_PWM_FREQUENCY:
             setPWMFrequency (ioCommand[1], ioCommand[2]);
           break;          
           case ANALOG_READ_POLLING_START:
             analogReadPin[analogReadPollingPinCount] = ioCommand[1]; // put on polling read list
             ++analogReadPollingPinCount;
           break;
           case ANALOG_READ_POLLING_STOP:
             removeAndShift(analogReadPin, analogReadPollingPinCount, ioCommand[1]);
           break;
           case DIGITAL_READ_POLLING_START:
             digitalReadPin[digitalReadPollingPinCount] = ioCommand[1]; // put on polling read list
             ++digitalReadPollingPinCount;
           break;
           case DIGITAL_READ_POLLING_STOP:
             removeAndShift(digitalReadPin, digitalReadPollingPinCount, ioCommand[1]);
           break;
        }

        //SoftwareServo::refresh();

        // reset buffer
        ioCommand[0] = -1;
        ioCommand[1] = -1;
        ioCommand[2] = -1;
        byteCount = 0;           
     
     } // if (byteCount > 2>   
  } // if (Serial.available () > 0)
  // handle serial data end

  // polling read - send data for pins which are currently in INPUT mode only AND whose state has changed
  for (int i  = 0; i < digitalReadPollingPinCount; ++i)
  {
    // read the pin
    readValue = digitalRead(digitalReadPin[i]);
      
    // if my value is different then last time - send it
    if (lastDigitalInputValue[digitalReadPin[i]] != readValue)
    {
      Serial.write(DIGITAL_VALUE);
      Serial.write(digitalReadPin[i]); // TODO - have to encode it to determine where it came from
      Serial.write(readValue); 
    }
    
    // set the last input value of this pin
    lastDigitalInputValue[digitalReadPin[i]] = readValue;
  }
    

  // polling read - send data for pins which are currently in INPUT mode only AND whose state has changed
  for (int i  = 0; i < analogReadPollingPinCount; ++i)
  {
    // read the pin
    readValue = analogRead(analogReadPin[i]);
      
    // if my value is different then last time - send it
    if (lastAnalogInputValue[analogReadPin[i]] != readValue)
    {
       
      sprintf(dbg,"%d%d%d\n");      
//      Serial.write(dbg);
//      Serial.write(ANALOG_VALUE);
//      Serial.write(analogReadPin[i]); // TODO - have to encode it to determine where it came from
//      Serial.write(readValue); 
    }
    // set the last input value of this pin
    lastAnalogInputValue[analogReadPin[i]] = readValue;
  }

}

